"""BDAP (JRC Big Data Analytics Platform) layer creation with minimal dependencies
(to create server-side inter.VectorLayer instances for vector display without using the client version of inter)."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Python imports
import ipyleaflet
from io import StringIO, BytesIO
import sys
import json
import os
import datetime
import requests
from PIL import Image, ImageDraw

# vois import
from vois.vuetify import textlist

# Local imports
import rasterAPI


#####################################################################################################################################################
# Notes on symbology:
#
# A symbol is a list of lists of items each having 3 elements: [SymbolizerName, KeyName, Value]
# Each list inside the symbol is mapped into a style (from style0 to style9), thus allowing for overlapped symbols
#
# Example:
# symbol = [
#             [
#                ["PolygonSymbolizer", "fill", '#ff0000'],
#                ["PolygonSymbolizer", "fill-opacity", 0.3],
#                ["LineSymbolizer", "stroke", "#010000"],
#                ["LineSymbolizer", "stroke-width", 2.0]
#             ]
# ]
#
# Example on how to manage symbology:
#
#    vlayer = vectorlayer.file('path to a .shp file', epsg=4326)
#    vlayer.symbologyClear()
#    vlayer.symbologyAdd(symbol=symbol)                              # Apply symbol to all features of the vectorlayer
#    vlayer.symbologyAdd(rule="[CNTR_CODE] = 'IT'", symbol=symbol)   # Apply symbol only to features that satisfy the rule on attributes
#                                                                    # See https://github.com/mapnik/mapnik/wiki/Filter for help on filter sintax
#    mapUtils.addLayer(m, vlayer.tileLayer(), name='Polygons')
#
#
# The static methos vectorlayer.symbolChange can be used to change a parametric symbol
#
# Example:
# symbol = [
#             [
#                ["PolygonSymbolizer", "fill", 'FILL-COLOR'],
#                ["PolygonSymbolizer", "fill-opacity", 0.3],
#                ["LineSymbolizer", "stroke", "#010000"],
#                ["LineSymbolizer", "stroke-width", 2.0]
#             ]
# ]
#
# s = vectorlayer.symbolChange(fillColor='red')
#
#####################################################################################################################################################


#####################################################################################################################################################
# Class vectorlayer to create server-side VectorLayer instances for vector display without using inter client library
# Manages vector datasets in files (shapefiles, geopackage, etc.), WKT strings and POSTGIS queries
#####################################################################################################################################################
class vectorlayer:
    
    # Initialization for vector files (shapefiles, geopackage, etc.)
    def __init__(self,
                 filepath='',
                 layer='',
                 epsg=4326,
                 proj='',              # To be used for projections that do not have an EPSG code (if not empty it is used instead of the passed epsg)
                 identify_fields=[]):  # List of names of field to display on identify operation

        self.isPostgis = False
        self.isWKT     = False
        
        self.filepath = filepath
        self.layer    = layer
        self.epsg     = epsg
        self.proj     = proj
        
        self._identify_fields = identify_fields
        
        
        # Store the procid (after a call to self.toLayer())
        self.procid = None
        
        # Symbology rules
        self.symbologyRules = []

        
    #####################################################################################################################################################
    # Initialization for a vector file (shapefile, geopackage, etc.)
    #####################################################################################################################################################
    @classmethod
    def file(cls,
             filepath,      # Path to the file (shapefile or geopackage, etc...)
             layer='',      # Name of the layer (for a shapefile leave it empty)
             epsg=4326,
             proj=''):      # To be used for projections that do not have an EPSG code (if not empty it is used instead of the passed epsg)
    
        instance = cls(filepath, layer, epsg, proj)
        return instance
    
    
    #####################################################################################################################################################
    # Initialization from a list of wkt strings
    #####################################################################################################################################################
    @classmethod
    def wkt(cls,
            wktlist,          # List of strings containing WKT of geospatial features in EPSG4326
            properties=[]):   # List of dictionaries containing the attributes of each of the feature (optional)
    
        instance = cls('', '', 4326, '')
        instance.isWKT      = True
        instance.wktlist    = wktlist
        instance.properties = properties
        return instance
    
    
    #####################################################################################################################################################
    # Initialization for a postGIS query
    #####################################################################################################################################################
    @classmethod
    def postgis(cls,
                host,
                port,
                dbname,
                user,
                password,
                query,
                epsg=4326,
                proj='',             # To be used for projections that do not have an EPSG code (if not empty it is used instead of the passed epsg)
                geomtype='polygon',
                geometry_field='',
                geometry_table='',
                extents=''):
        
        instance = cls()

        instance.isPostgis = True
        instance.isWKT     = False
        
        instance.postgis_host           = host
        instance.postgis_port           = port
        instance.postgis_dbname         = dbname
        instance.postgis_user           = user
        instance.postgis_password       = password
        instance.postgis_query          = query
        instance.postgis_epsg           = epsg
        instance.postgis_proj           = proj
        instance.postgis_geomtype       = geomtype
        instance.postgis_geometry_field = geometry_field
        instance.postgis_geometry_table = geometry_table
        instance.postgis_extents        = extents

        return instance                

    
    #####################################################################################################################################################
    # Symbology management
    #####################################################################################################################################################
    
    # Symbology reset (remove all symbol rules)
    def __symbologyReset(self):
        self.symbologyRules = []
        
    # Symbology remove (add a rule to remove the default symbology)
    def __symbologyRemove(self, styleName='default', ruleName='all'):
        self.symbologyRules.append(RuleModify(changeType=False, styleName=styleName, ruleName=ruleName))
        
    # Symbology remove (add a rule to change the symbology)
    def __symbologySet(self, styleName='default', ruleName='all', symbolizerName='', keyName='', value=''):
        self.symbologyRules.append(RuleModify(changeType=True, styleName=styleName, ruleName=ruleName, symbolizerName=symbolizerName, keyName=keyName, value=value))

    
    # Remove all default symbology and all symbols added)
    def symbologyClear(self, maxstyle=0):
        
        self.__symbologyReset()
        self.__symbologyRemove(styleName='default', ruleName='all')
        
        # Clear the style from 0 to 
        for stylelayer in range(maxstyle+1):
            self.__symbologyRemove(styleName="style%d"%stylelayer, ruleName='all')

            
    # Apply a symbol to a subset of the features filtered by a rule ('all' applies to all features, "[attrib] = 'value'" only to a subset of the features. See https://github.com/mapnik/mapnik/wiki/Filter for filter sintax)
    def symbologyAdd(self, rule='all', symbol=[]):
        stylelayer = 0
        for layer in symbol:
            style = "style%d"%stylelayer
            #style = "default"
            for member in layer:
                symbolizer,attribute,value = member
                self.__symbologySet(style, rule, symbolizer, attribute, str(value))
            stylelayer += 1
                
    
    # Change color and other properties of a symbol and returns the modified symbol
    @staticmethod
    def symbolChange(symbol, color='#ff0000', fillColor='#ff0000', fillOpacity=1.0, strokeColor='#ffff00', strokeWidth=0.5, scalemin=None, scalemax=None):
        newsymbol = []
        for layer in symbol:
            newlayer = []
            for member in layer:
                symbolizer,attribute,value = member

                if value == 'COLOR':
                    value = color

                if value == 'FILL-COLOR':
                    value = fillColor

                if value == 'FILL-OPACITY':
                    value = fillOpacity

                if value == 'STROKE-COLOR':
                    value = strokeColor

                if value == 'STROKE-WIDTH':
                    value = strokeWidth

                if value == 'SCALE-MIN':
                    value = scalemin

                if value == 'SCALE-MAX':
                    value = scalemax

                if not value is None:
                    newlayer.append((symbolizer,attribute,value))

            newsymbol.append(newlayer)

        return newsymbol

    
    #####################################################################################################################################################
    # Print
    #####################################################################################################################################################
    
    # Representation
    def __repr__(self):
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        self.print()
        sys.stdout = old_stdout
        return mystdout.getvalue()
        
        
    # Print info on instance    
    def print(self):
        if self.isPostgis:
            print("BDAP vector layer POSTGIS:")
            print("   procid:         %s"%str(self.procid))
            print("   host:           %s"%self.postgis_host)
            print("   port:           %d"%self.postgis_port)
            print("   dbname:         %s"%self.postgis_dbname)
            print("   user:           %s"%self.postgis_user)
            print("   password:       %s"%self.postgis_password)
            print("   query:          %s"%self.postgis_query)
            print("   epsg:           %d"%self.epsg)
            print("   proj:           %s"%self.proj)
            print("   geomtype:       %s"%self.postgis_geomtype)
            print("   geometry_field: %s"%self.postgis_geometry_field)
            print("   geometry_table: %s"%self.postgis_geometry_table)
            print("   extents:        %s"%self.postgis_extents)
        elif self.isWKT:
            print("BDAP vector layer WKT:")
            print("   wktlist:        %s"%str(self.wktlist))
            print("   properties:     %s"%str(self.properties))
        else:
            print("BDAP vector layer FILE:")
            print("   procid:         %s"%str(self.procid))
            print("   filepath:       %s"%self.filepath)
            print("   layer:          %s"%self.layer)
            print("   epsg:           %d"%self.epsg)
            print("   proj:           %s"%self.proj)
            
        print("   symbology:")
        for rule in self.symbologyRules:
            print("       %s"%rule.print())
    

    
    #####################################################################################################################################################
    # Identify methods
    #####################################################################################################################################################
    
    # Identify: returns a string
    def identify(self, lon, lat, zoom):
        while lon < -180.0: lon += 360.0
        while lon >  180.0: lon -= 360.0
        url = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view/?IDENTIFYEX=1&vl=%s&x=%f&y=%f&epsg=%d&zoom=%d' % (self.toLayer(), lon, lat, 4326, int(zoom))
        #print(url)
        response = requests.get(url)
        bio = BytesIO(response.content)
        s = bio.read()
        svalue = s.decode("utf-8").replace('Values = ','').replace('Value = ','')
        return svalue
        

    # onclick called by a Map.Map instance
    def onclick(self, m, lon, lat, zoom):
        res = self.identify(lon, lat, zoom)
        if not res is None and len(res) > 0:
            if '#' in res:
                descriptions = self._identify_fields
                values       = res.split('#')
            else:
                if ', ' in res:
                    attribs = res.split(', ')
                    
                    descriptions = []
                    values       = []
                    for attrib in attribs:
                        s = attrib.split('=')
                        if len(s) >= 2:
                            descriptions.append(s[0])
                            values.append(s[1])
                else:
                    descriptions = ['Attributes']
                    values       = [res]

            t = textlist.textlist(descriptions, values,
                                  titlefontsize=10,
                                  textfontsize=11,
                                  titlecolumn=4,
                                  textcolumn=8,
                                  titlecolor='#000000',
                                  textcolor='#000000',
                                  lineheightfactor=1.1)

            t.card.width = '180px'
            popup = ipyleaflet.Popup(location=[lat,lon], child=t.draw(), auto_pan=False, close_button=True, auto_close=True, close_on_escape_key=True)
            m.add_layer(popup)
            
            
    #####################################################################################################################################################
    # Properties
    #####################################################################################################################################################

    @property
    def identify_fields(self):
        return self._identify_fields
        
    @identify_fields.setter
    def identify_fields(self, listofattributes):
        self._identify_fields = listofattributes
                
                
    #####################################################################################################################################################
    # Create an ipyleaflet.TileLayer
    #####################################################################################################################################################
    
    # Returns an instance of ipyleaflet.TileLayer
    def tileLayer(self, max_zoom=22):
        url = self.tileUrl()
        if not url is None:
            return ipyleaflet.TileLayer(url=url, max_zoom=max_zoom, max_native_zoom=max_zoom)

        
    #####################################################################################################################################################
    # Internal functions
    #####################################################################################################################################################
    
    # Returns the url to display the layer
    def tileUrl(self):
        procid = self.toLayer()
        if not procid is None:
            tileBaseUrl = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view'
            return tileBaseUrl + "?x={x}&y={y}&z={z}&procid=%s" % procid

        
    # Save the layer in Redis and returns the procid
    def toLayer(self):
        j = self.toJson()
        strjson = json.dumps(j)
        self.procid = rasterAPI.saveLayer(strjson)
        return self.procid

    
    # Return the JSON representation of the vector layer
    def toJson(self):
        filelinkproj = "+init=epsg:%d"%self.epsg
        if self.proj is not None and len(self.proj) > 0:
            filelinkproj = self.proj
        
        filelinklayer = self.layer
        if len(filelinklayer) == 0:
            path, file = os.path.split(self.filepath)
            filelinklayer = file.rsplit('.', 1)[0]
        
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        j = {'AbsolutePath': '',
             'Collection': 1002,
             'CustomXML': '',
             'Description': '',
             'HeatmapMode': 0,
             'HeatmapQuery': '',
             'HeatmapRadius': 0,
             'HeatmapWeightField': '',
             'HeatmapWeightMax': 1000000,
             'HeatmapWeightMin': -1000000,
             'IdentifyAll': 0,
             'IdentifyDigits': -1,
             'IdentifyField': '',
             'IdentifyFilter': '',
             'IdentifySeparator': '#',
             'IdentifySortField': '',
             'Name': 'wkt',
             'POSTGIS_dbname': '',
             'POSTGIS_epsg': 4326,
             'POSTGIS_extents': '',
             'POSTGIS_geometry_field': '',
             'POSTGIS_geometry_table': '',
             'POSTGIS_geomtype': 'Polygon',
             'POSTGIS_host': '',
             'POSTGIS_password': '',
             'POSTGIS_port': 0,
             'POSTGIS_query': '',
             'POSTGIS_user': '',
             'Raster_XML': '',
             'Raster_band': 1,
             'Raster_colors': '#000000,#ffffff',
             'Raster_epsgcode': 4326,
             'Raster_file': '',
             'Raster_interpolate': 'near',
             'Raster_nodata': 0,
             'Raster_scalemax': 255,
             'Raster_scalemin': 0,
             'ScaleResolution': 1,
             'filelinklayer': '',
             'filelinkpath': '',
             'filelinkproj': '',
             'joins': None,
             'modify': None,
             'opacity': 255,
             'properties': None,
             'wkt': None}    
        
        # Set the identify fields
        if isinstance(self._identify_fields, list):
            if len(self._identify_fields) > 0:
                j['IdentifyField'] = ' '.join(self._identify_fields)
                
        # Rules to modify the symbology
        j["modify"] = [x.toJson() for x in self.symbologyRules]
        
        # Add specific settings of the three formats
        if self.isPostgis:
            j["Name"]                   = "postgis"
            j["POSTGIS_host"]           = self.postgis_host
            j["POSTGIS_port"]           = self.postgis_port
            j["POSTGIS_dbname"]         = self.postgis_dbname
            j["POSTGIS_user"]           = self.postgis_user
            j["POSTGIS_password"]       = self.postgis_password
            j["POSTGIS_query"]          = self.postgis_query
            j["POSTGIS_epsg"]           = self.postgis_epsg
            j["POSTGIS_proj"]           = self.postgis_proj
            j["POSTGIS_geomtype"]       = self.postgis_geomtype
            j["POSTGIS_geometry_field"] = self.postgis_geometry_field
            j["POSTGIS_geometry_table"] = self.postgis_geometry_table
            j["POSTGIS_extents"]        = self.postgis_extents
        elif self.isWKT:
            j["Name"] = "wkt"
            j["wkt"]  = self.wktlist
            while len(self.properties) < len(self.wktlist): self.properties.append({})
            j["properties"] = [str(x).replace("'",'"') for x in self.properties]
        else:
            j["Name"]          = "wkt"
            j["filelinklayer"] = filelinklayer
            j["filelinkpath"]  = self.filepath
            j["filelinkproj"]  = filelinkproj
        
        return j

    
    
#####################################################################################################################################################
# Generate an image from a symbol
#####################################################################################################################################################
def symbol2Image(symbol=[], size=1, feature='Point', clipdimension=999, showborder=False):

    # Symbols dimension in pixels
    SMALL_SYMBOLS_DIMENSION  = 30
    MEDIUM_SYMBOLS_DIMENSION = 80
    LARGE_SYMBOLS_DIMENSION  = 256
    
    doclip = False
    if feature == 'Line':
        if size >= 3:    wkt = 'LINESTRING (-170 82, -100 55, -60 70, -10 38)'
        elif size == 2:  wkt = 'LINESTRING (-175 83, -158 81, -148 83, -129 81)'
        else:            wkt = 'LINESTRING (-177 84.45, -171 83.9, -167.4 84.25, -161 83.75)'
    elif feature == 'Polygon':
        if size >= 3:    wkt = 'POLYGON ((-170 83.85, -170 10, -10 10, -10 83.85, -170 83.85))'
        elif size == 2:  wkt = 'POLYGON ((-175 84.5, -175 78, -128.5 78, -128.5 84.5, -175 84.5))'
        else:            wkt = 'POLYGON ((-178 84.85, -178 83.2, -160.5 83.2, -160.5 84.85, -178 84.85))'
    else:
        if size >= 3:    wkt = 'POINT (-90 65)'
        elif size == 2:  wkt = 'POINT (-152 82)'
        else:            wkt = 'POINT (-169.52 84.05)'

    if size >= 3:
        if clipdimension < LARGE_SYMBOLS_DIMENSION:
            doclip = True
    elif size == 2:
        if clipdimension < MEDIUM_SYMBOLS_DIMENSION:
            doclip = True
    else:
        if clipdimension < SMALL_SYMBOLS_DIMENSION:
            doclip = True

    vlayer = vectorlayer.wkt([wkt])
    vlayer.symbologyClear()
    vlayer.symbologyAdd(symbol=symbol)
    #print(vlayer.toJson())

    url = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view?x=0&y=0&z=1&procid=%s' % vlayer.toLayer()
    response = requests.get(url)
    
    if len(response.content) > 5 and response.content[0] == 137 and response.content[1] == 80 and response.content[2] == 78 and response.content[3] == 71 and response.content[4] == 13:
        img = Image.open(BytesIO(response.content))
        if size >= 3:
            img = img.crop((0, 0, LARGE_SYMBOLS_DIMENSION, LARGE_SYMBOLS_DIMENSION))
        elif size == 2:
            img = img.crop((0, 0, MEDIUM_SYMBOLS_DIMENSION, MEDIUM_SYMBOLS_DIMENSION))
        else:
            img = img.crop((0, 0, SMALL_SYMBOLS_DIMENSION, SMALL_SYMBOLS_DIMENSION))

        if doclip:
            s = img.size
            cx = s[0]/2
            cy = s[1]/2
            img = img.crop((cx-clipdimension/2, cy-clipdimension/2, cx+clipdimension/2, cy+clipdimension/2))
    else:
        print('URL with errors:',url)
        if size >= 3:    img = Image.new("RGB", (LARGE_SYMBOLS_DIMENSION,  LARGE_SYMBOLS_DIMENSION),  (255, 255, 255))
        elif size == 2:  img = Image.new("RGB", (MEDIUM_SYMBOLS_DIMENSION, MEDIUM_SYMBOLS_DIMENSION), (255, 255, 255))
        else:            img = Image.new("RGB", (SMALL_SYMBOLS_DIMENSION,  SMALL_SYMBOLS_DIMENSION),  (255, 255, 255))
        draw = ImageDraw.Draw(img)
        draw.text((0, 0),"Error",(0,0,0))

    # Add a thin black border
    if showborder:
        draw = ImageDraw.Draw(img)
        s = img.size
        draw.rectangle(((0, 0), (s[0]-1, s[1]-1)), outline='black')

    return img
    
        
#####################################################################################################################################################
# RuleModify class: modification to a symbology rule (see interapro VectorLayer.cpp)
#####################################################################################################################################################
class RuleModify():
    
    # Initialization
    def __init__(self,
                 changeType,   # If true it is a change rule, if False a remove rule
                 styleName,
                 ruleName,
                 symbolizerName='',
                 keyName='',
                 value=''):
    
        self.changeType     = changeType
        self.styleName      = styleName
        self.ruleName       = ruleName
        self.symbolizerName = symbolizerName
        self.keyName        = keyName
        self.value          = value

        
    # Return a string showing the rule members
    def print(self):
        if self.changeType:
            return 'type: CHANGE    style: %-10s    rule: %-20s    symbolizer: %-20s     key: %-16s    value: %-20s'%(self.styleName, self.ruleName, self.symbolizerName, self.keyName, self.value)
        else:
            return 'type: REMOVE    style: %-10s    rule: %-20s'%(self.styleName, self.ruleName)
        
        
    # Serialization to Json
    def toJson(self):
        jsonobj = {}

        if self.changeType:
            jsonobj["type"] = 0     # Change rule
        else:
            jsonobj["type"] = 1     # Remove rule
            
        jsonobj["style"] = self.styleName
        jsonobj["rule"]  = self.ruleName
        
        if self.changeType:
            jsonobj["symbolizer"] = self.symbolizerName
            jsonobj["key"]        = self.keyName
            jsonobj["value"]      = self.value
    
        return jsonobj
    


"""Map class"""
# Author(s): Davide.De-Marchi@ec.europa.eu, Edoardo.Ramalli@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
from ipywidgets import widgets, HTML, Layout, CallbackDispatcher
import ipyleaflet
from ipyleaflet import SearchControl, ScaleControl, FullScreenControl, WidgetControl
import ipyvuetify as v

# Vois imports
from vois.vuetify import settings, toggle, switch
from vois.geo import mapUtils
from vois.templates import PageConfigurator

# Name of layers
LAYERNAME_BACKGROUND  = 'Background'
LAYERNAME_LABELS      = 'Labels'


#####################################################################################################################################################
# Map class
#####################################################################################################################################################
class Map(ipyleaflet.Map):


    # Initialization
    def __init__(self,
                 width='100%',
                 height='600px',
                 center=[50, 12],
                 zoom=5,
                 show_fullscreen=True,    # Show or Hide the fullscreen control
                 show_search=True,        # Show or Hide the Search control
                 show_scale=True,         # Show or Hide the Scale control
                 show_coordinates=True,   # Show or hide the Coordinates control
                 show_overview=False,     # Show or hide the Overview control
                 show_basemaps=True,      # Show or hide the Basemaps toggle control
                 color_first=None,        # Main color
                 color_second=None,       # Secondary color
                 dark=None,               # Dark flag
                 basemapindex=0,          # Initial basemap index (0=EC, 1=Esri, 2=Google)
                 onclick=None,            # Callback function to call on click (receives as parameter: map, lon, lat, zoom)
                 **kwargs):
        
        self._width            = width
        self._height           = height
        self._show_fullscreen  = show_fullscreen
        self._show_search      = show_search
        self._show_scale       = show_scale
        self._show_coordinates = show_coordinates
        self._show_overview    = show_overview
        self._show_basemaps    = show_basemaps
        self._onclick          = onclick
        
        self._color_first = color_first
        if self._color_first is None:
            self._color_first = settings.color_first
        
        self._color_second = color_second
        if self._color_second is None:
            self._color_second = settings.color_second
            
        self._dark = dark
        if self._dark is None:
            self._dark = settings.dark_mode
            
        self._basemapindex = basemapindex
        
        # Initial center and zoom of the map
        self.center = center
        self.zoom   = zoom

        # Card containing the widgets to configure the map appearance
        self.configure_card = None
        self.s1 = self.s2 = self.s3 = self.s4 = self.s5 = self.s6 = None
        self.update_properties = True
        
        # Map widget
        super().__init__(max_zoom=21, center=self.center, zoom=self.zoom, scroll_wheel_zoom=True, 
                         basemap=mapUtils.EmptyBasemap(), attribution_control=False, layout=Layout(width=self._width, height=self._height, margin='0px 0px 0px 0px'),
                         **kwargs)
        
        mapUtils.addLayer(self, mapUtils.OSM_EC(),      LAYERNAME_BACKGROUND)
        mapUtils.addLayer(self, mapUtils.CartoLabels(), LAYERNAME_LABELS)
        layer = mapUtils.getLayer(self, LAYERNAME_LABELS)
        layer.opacity = 0.0
        self.onSelectBasemap(self._basemapindex)
        
        self.toggleBasemap   = None
        self.cardCoordinates = None
        
        # Add FullScreen control
        self.show_fullscreen = self._show_fullscreen

        # Add Search control
        self.show_search = self._show_search

        # Add Scale control
        self.show_scale = self._show_scale

        # Add widget control to select basemaps
        self.show_basemaps = self._show_basemaps
        
        # Add overview
        self.show_overview = self._show_overview
        
        # Add widget control to display geographic coordinates at mouse move
        self.show_coordinates = self._show_coordinates
        
        # Manage interaction events
        self._interaction_callbacks = CallbackDispatcher()
        self.on_interaction(self.handleMapInteraction)
        

    # Fake "draw" method
    def draw(self):
        return self
      
    
    # Remove all layers from map
    def clear(self):
        mapUtils.clear(self)
    
    
    # Add a ipyleaflet.TileLayer to the map
    def addLayer(self, tileLayer, name=None, opacity=1.0):
        
        # if no name is passed, generate a layer name
        if name is None:
            count = len(self.layers)
            name = "layer%d"%(count+1)
            
        # In case a vectorlayer or rasterlayer is passed: call .tileLayer() to get the ipyleaflet.TileLayer instance!
        if isinstance(tileLayer, ipyleaflet.TileLayer):
            mapUtils.addLayer(self, tileLayer, name=name, opacity=opacity)
        else:
            mapUtils.addLayer(self, tileLayer.tileLayer(), name=name, opacity=opacity)
        
        
    # Manage all user interaction on the map
    def handleMapInteraction(self, **kwargs):
        
        if self._show_coordinates:
            if kwargs.get('type') == 'mousemove':
                lon = kwargs.get('coordinates')[1]
                lat = kwargs.get('coordinates')[0]
                self.cardCoordinates.children = [v.Html(tag='div', children=[' %.4f° N, %.4f° E'%(lat,lon)], style_='color: black;', class_='pa-0 ma-0 ml-1 mr-1')]
        
        if self._onclick is not None:
            if kwargs.get('type') == 'click':
                lon = kwargs.get('coordinates')[1]
                lat = kwargs.get('coordinates')[0]
                self._onclick(self, lon, lat, int(self.zoom))
        

    # Selection of a basemap
    def onSelectBasemap(self, index):
        layer = mapUtils.getLayer(self, LAYERNAME_LABELS)

        self._basemapindex = max(0, min(2, index))
        
        if self._basemapindex == 1:
            basemaplayer = mapUtils.EsriWorldImagery()
            layer.opacity = 1.0
        elif self._basemapindex == 2:
            basemaplayer = mapUtils.GoogleHybrid()
            layer.opacity = 0.0
        else:
            basemaplayer = mapUtils.OSM_EC()
            layer.opacity = 0.0

        mapUtils.addLayer(self, basemaplayer, LAYERNAME_BACKGROUND)

        
    # Change of configure widgets
    def change_fullscreen(self, flag): 
        if self.update_properties:
            self.show_fullscreen = flag
            
    def change_coordinates(self, flag):
        if self.update_properties:
            self.show_coordinates = flag
            
    def change_search(self, flag):
        if self.update_properties:
            self.show_search = flag
            
    def change_scale(self, flag):
        if self.update_properties:
            self.show_scale = flag
            
    def change_basemaps(self, flag):
        if self.update_properties:
            self.show_basemaps = flag
            
    def change_overview(self, flag):
        if self.update_properties:
            self.show_overview = flag
        
        
    # Create a card instance to configure the Map appearance and returns it
    def configure(self):
        if self.configure_card is None:
            self.configure_card = v.Card(flat=True, class_='pa-2 ma-0')
            
            self.s1 = switch.switch(self.show_fullscreen,  'Add the Fullscreen control',  inset=True, dense=True, onchange=self.change_fullscreen, color=self._color_first)
            self.s2 = switch.switch(self.show_coordinates, 'Add the Coordinates control', inset=True, dense=True, onchange=self.change_coordinates, color=self._color_first)
            self.s3 = switch.switch(self.show_search,      'Add the Search control',      inset=True, dense=True, onchange=self.change_search, color=self._color_first)
            self.s4 = switch.switch(self.show_scale,       'Add the Scale control',       inset=True, dense=True, onchange=self.change_scale, color=self._color_first)
            self.s5 = switch.switch(self.show_basemaps,    'Add the Basemaps control',    inset=True, dense=True, onchange=self.change_basemaps, color=self._color_first)
            self.s6 = switch.switch(self.show_overview,    'Add the Overview control',    inset=True, dense=True, onchange=self.change_overview, color=self._color_first)
            
            self.configure_card.children = [widgets.VBox([PageConfigurator.label('Map', color='black'), self.s1.draw(), self.s2.draw(), self.s3.draw(), self.s4.draw(), self.s5.draw(), self.s6.draw()])]
            
        return self.configure_card


    @property
    def content(self):
        return 'Map'
        
    @content.setter
    def content(self, c):
        pass
    
    
    @property
    def width(self):
        return self._width
        
    @width.setter
    def width(self, w):
        self._width = w
        self.layout.width = self._width

        
    @property
    def height(self):
        return self._height
        
    @height.setter
    def height(self, h):
        self._height = h
        self.layout.height = self._height

        
    @property
    def show_fullscreen(self):
        return self._show_fullscreen
        
    @show_fullscreen.setter
    def show_fullscreen(self, flag):
        self._show_fullscreen = flag

        if self.s1 is not None:
            self.update_properties = False
            self.s1.value = flag
            self.update_properties = True

        for control in self.controls:
            if isinstance(control, ipyleaflet.leaflet.FullScreenControl):
                self.remove(control)
                
        if self._show_fullscreen:
            self.add(FullScreenControl(position="topright"))


    @property
    def show_coordinates(self):
        return self._show_coordinates
        
    @show_coordinates.setter
    def show_coordinates(self, flag):
        self._show_coordinates = flag

        if self.s2 is not None:
            self.update_properties = False
            self.s2.value = flag
            self.update_properties = True
        
        mapUtils.removeCoordinates(self)
                
        if self._show_coordinates:
            self.cardCoordinates = mapUtils.getCoordinatesCard(self)

            
    @property
    def show_search(self):
        return self._show_search
        
    @show_search.setter
    def show_search(self, flag):
        self._show_search = flag

        if self.s3 is not None:
            self.update_properties = False
            self.s3.value = flag
            self.update_properties = True

        for control in self.controls:
            if isinstance(control, ipyleaflet.leaflet.SearchControl):
                self.remove(control)
                
        if self._show_search:
            self.add(SearchControl(position="topleft",url='https://nominatim.openstreetmap.org/search?format=json&q={s}',zoom=12))


    @property
    def show_scale(self):
        return self._show_scale
        
    @show_scale.setter
    def show_scale(self, flag):
        self._show_scale = flag

        if self.s4 is not None:
            self.update_properties = False
            self.s4.value = flag
            self.update_properties = True

        for control in self.controls:
            if isinstance(control, ipyleaflet.leaflet.ScaleControl):
                self.remove(control)
                
        if self._show_scale:
            self.add(ScaleControl(position='topleft'))


    @property
    def show_basemaps(self):
        return self._show_basemaps
        
    @show_basemaps.setter
    def show_basemaps(self, flag):
        self._show_basemaps = flag

        if self.s5 is not None:
            self.update_properties = False
            self.s5.value = flag
            self.update_properties = True

        mapUtils.removeCardByName(self, 'Basemaps')
                
        if self._show_basemaps:
            c = mapUtils.getCardByName(self, 'Basemaps', position='bottomright')
            c.tile = True
            self.toggleBasemap = toggle.toggle(self._basemapindex,
                                               ['Gisco', 'Esri', 'Google'],
                                               tooltips=['Select EC Gisco roadmap as background layer', 'Select ESRI WorldImagery as background layer', 'Select GOOGLE Satellite as background layer'],
                                               colorselected=self._color_first, colorunselected=self._color_second, rounded=False,
                                               dark=self._dark, onchange=self.onSelectBasemap, row=True, width=70, justify='start', paddingrow=0, tile=True)
            c.children = [self.toggleBasemap.draw()]


    @property
    def show_overview(self):
        return self._show_overview
        
    @show_overview.setter
    def show_overview(self, flag):
        self._show_overview = flag
        mapUtils.removeOverview(self)
        
        if self.s6 is not None:
            self.update_properties = False
            self.s6.value = flag
            self.update_properties = True
                
        if self._show_overview:
            mapUtils.addOverview(self, position='bottomleft')


    @property
    def color_first(self):
        return self._color_first
        
    @color_first.setter
    def color_first(self, color):
        self._color_first = color

        if self.toggleBasemap is not None:
            self.toggleBasemap.colorselected = self._color_first

        if self.configure_card is not None:
            self.s1.color = self._color_first
            self.s2.color = self._color_first
            self.s3.color = self._color_first
            self.s4.color = self._color_first
            self.s5.color = self._color_first
            self.s6.color = self._color_first


    @property
    def color_second(self):
        return self._color_second
        
    @color_second.setter
    def color_second(self, color):
        self._color_second = color

        if self.toggleBasemap is not None:
            self.toggleBasemap.colorunselected = self._color_second


    @property
    def dark(self):
        return self._dark
        
    @dark.setter
    def dark(self, flag):
        self._dark = flag

        if self.toggleBasemap is not None:
            self.toggleBasemap.dark = self._dark
            
            
    @property
    def basemapindex(self):
        return self._basemapindex
        
    @basemapindex.setter
    def basemapindex(self, index):
        self.onSelectBasemap(index)

        if self.toggleBasemap is not None:
            self.toggleBasemap.value = self._basemapindex
            
            
    @property
    def state(self):
        return {x: getattr(self, x) for x in ['content',
                                              #'width',      # Will inherit from content!!!
                                              #'height',
                                              'show_fullscreen',
                                              'show_coordinates',
                                              'show_search',
                                              'show_scale',
                                              'show_basemaps',
                                              'show_overview',
                                              'color_first',
                                              'color_second',
                                              'dark',
                                              'basemapindex',
                                              'center',
                                              'zoom'
                                             ]}
        
    @state.setter
    def state(self, statusdict):
        for key, value in statusdict.items():
            setattr(self, key, value)
            
            
    @property
    def onclick(self):
        return self._onclick
        
    @onclick.setter
    def onclick(self, callback):
        self._onclick = callback
            


"""Template page with left and bottom panels"""
# Author(s): Davide.De-Marchi@ec.europa.eu, Edoardo.Ramalli@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
from ipywidgets import widgets, HTML, Layout, CallbackDispatcher
import ipyleaflet
from ipyleaflet import SearchControl, ScaleControl, FullScreenControl, WidgetControl
import ipyvuetify as v

# Vois imports
from vois.vuetify import settings, toggle, page
from vois.templates import dynamicButton, Content

# Panels dimensioning
LEFT_WIDTH    = 400      # Width  in pixels of the left bar
BOTTOM_HEIGHT = 240      # Height in pixels of the bottom bar

# Name of layers
LAYERNAME_BACKGROUND  = 'Background'
LAYERNAME_LABELS      = 'Labels'

                
#####################################################################################################################################################
# Template page with left and bottom panels
#####################################################################################################################################################
class template2panels(page.page):

    
    # Initialization
    def __init__(self, output, onclose=None, leftWidth=LEFT_WIDTH, bottomHeight=BOTTOM_HEIGHT, **kwargs):
        super().__init__('Demo', 'Geospatial page with left and bottom panel', output, onclose=onclose, copyrighttext='European Commission - Joint Research Centre', **kwargs)

        # Initialize member variables
        self.init_leftWidth    = self.leftWidth    = leftWidth
        self.init_bottomHeight = self.bottomHeight = bottomHeight
   

    #################################################################################################################################################
    # Create the page and returns the card widget where the content of the page must be displayed
    #################################################################################################################################################
    def create(self):
        super().create()
        
        # Cards for the panels
        st = 'border-radius: 0px; border-color: %s; border-width: 1px; overflow: hidden;'%settings.color_first
        self.main_width  = 'calc(100vw - %dpx)'%self.leftWidth
        self.main_height = 'calc(%s - %dpx)'%(self.height,self.bottomHeight)
        self.cardLeft   = v.Card(flat=True, style_=st, outlined=True, width=self.leftWidth, min_width=self.leftWidth, max_width=self.leftWidth, height=self.height)
        self.cardBottom = v.Card(flat=True, style_=st + ' border-left-width: 0px; border-top-width: 0px;', outlined=True, width=self.main_width, height=self.bottomHeight, min_height=self.bottomHeight)
        self.cardMain   = v.Card(flat=True, style_=st + ' border-left-width: 0px;', outlined=True, width=self.main_width, height=self.main_height)
        
        # DynamicButtons to open/close the left and bottom panels
        self.dynbLeft   = dynamicButton.dynamicButton(x1='%dpx'%(self.leftWidth-45), y1='64px', x2='48px', y2='64px', onclick1=self.leftClose, onclick2=self.leftOpen)
        self.dynbBottom = dynamicButton.dynamicButton(x1='calc(100vw - 45px)', y1='calc(100vh - %dpx)'%(BOTTOM_HEIGHT+35), x2='calc(100vw - 45px)', y2='calc(100vh - 40px)',
                                                      icon1='mdi-menu-down', icon2='mdi-menu-up', onclick1=self.bottomClose, onclick2=self.bottomOpen)
        
        # Creation of the contents for the panels
        self.createMain()
        self.createLeft()
        self.createBottom()
        
        # Compose the panels
        self.card.children = [self.dynbLeft.draw(),
                              self.dynbBottom.draw(),
                              widgets.HBox([
                                  self.cardLeft,
                                  widgets.VBox([
                                      self.cardMain,
                                      self.cardBottom
                                  ])
                              ])
                             ]
        
        return self.card
    

    @property
    def titlecolor(self):
        return self._titlecolor
        
    @titlecolor.setter
    def titlecolor(self, color):
        page.page.titlecolor.fset(self, color)   # call super() property setter

        # Set the color of the panels outline!
        st = 'border-radius: 0px; border-color: %s; border-width: 1px; overflow: hidden;'%color
        self.cardLeft.style_   = st
        self.cardBottom.style_ = st + ' border-left-width: 0px; border-top-width: 0px;'
        self.cardMain.style_   = st + ' border-left-width: 0px;'

        # Set the color of the dynamicButton
        self.dynbLeft.color   = color
        self.dynbBottom.color = color
        
        # Set the color first
        self.content.color_first = color
        
                
    @property
    def footercolor(self):
        return self._footercolor
        
    @footercolor.setter
    def footercolor(self, color):
        page.page.footercolor.fset(self, color)   # call super() property setter

        # Set the color second
        self.content.color_second = color
                
        
    @property
    def titleheight(self):
        return self._titleheight
        
    @titleheight.setter
    def titleheight(self, height):
        page.page.titleheight.fset(self, height)   # call super() property setter

        self.main_height = 'calc(%s - %dpx)'%(self.height,self.bottomHeight)
        self.cardLeft.height = self.height
        self.cardMain.height = self.main_height
        
        self.content.height = 'calc(%s - %fpx)'%(self.height,self.bottomHeight)
        
        d = self._titleheight - 54
        newy = '%dpx'%(64+d)
        self.dynbLeft.y1 = self.dynbLeft.y2 = newy
        self.dynbLeft.setpos()
        

    @property
    def footerheight(self):
        return self._footerheight
        
    @footerheight.setter
    def footerheight(self, height):
        page.page.footerheight.fset(self, height)   # call super() property setter

        self.main_height = 'calc(%s - %dpx)'%(self.height,self.bottomHeight)
        self.cardLeft.height = self.height
        self.cardMain.height = self.main_height
        
        self.content.height = 'calc(%s - %fpx)'%(self.height,self.bottomHeight)
        
        d = self._titleheight - 54
        newy = '%dpx'%(64+d)
        self.dynbLeft.y1 = self.dynbLeft.y2 = newy
        self.dynbLeft.setpos()
        
        d = self._footerheight - 30
        self.dynbBottom.y1 = 'calc(100vh - %dpx)'%(BOTTOM_HEIGHT+35+d)
        self.dynbBottom.y2 = 'calc(100vh - %dpx)'%(40+d)
        self.dynbBottom.setpos()
        
        
    # Create the content of the left panel
    def createLeft(self):
        pass
    
    # Create the content of the bottom panel
    def createBottom(self):
        pass
        
    # Create the content of the Main panel
    def createMain(self):

        # Create the content instance
        self.content = Content.Content(output=self.output, width='calc(100vw - %dpx)'%self.leftWidth, height='calc(%s - %fpx)'%(self.height,self.bottomHeight), onStateChanged=self.onStateChanged)
        
        # Display the map inside the main card
        self.cardMain.children = [self.content]
    
    
    # Display the configuration GUI
    def configure(self):
        return self.content.configure()
        
        
    # Called when the state changes
    def onStateChanged(self):
        pass

    
    #################################################################################################################################################
    # Manage the opening/closing of the dynamic panels (left and bottom)
    #################################################################################################################################################
        
    # Close the left panel
    def leftClose(self):
        self.leftWidth = 0
        self.main_width  = '100vw'
        self.cardLeft.width     = self.leftWidth
        self.cardLeft.min_width = self.leftWidth
        self.cardLeft.max_width = self.leftWidth
        
        self.cardBottom.width = '100vw'
        self.cardMain.width   = self.main_width
        self.content.width    = self.main_width
        

    # Open the left panel
    def leftOpen(self):
        self.leftWidth = self.init_leftWidth
        self.main_width = 'calc(100vw - %dpx)'%self.leftWidth
        self.cardLeft.width     = self.leftWidth
        self.cardLeft.min_width = self.leftWidth
        self.cardLeft.max_width = self.leftWidth
        
        self.cardBottom.width = self.main_width
        self.cardMain.width   = self.main_width
        self.content.width    = self.main_width
        

    # Close the bottom panel
    def bottomClose(self):
        self.bottomHeight = 0
        self.cardBottom.height     = self.bottomHeight
        self.cardBottom.min_height = self.bottomHeight
        
        self.cardMain.height = self.height
        self.content.height  = self.height
        

    # Open the bottom panel
    def bottomOpen(self):
        self.bottomHeight = self.init_bottomHeight
        self.cardBottom.height     = self.bottomHeight
        self.cardBottom.min_height = self.bottomHeight
        
        self.cardMain.height = 'calc(%s - %dpx)'%(self.height,self.bottomHeight)
        self.content.height  = 'calc(%s - %fpx)'%(self.height,self.bottomHeight)
        



"""mainPage Configurator"""
# Author(s): Davide.De-Marchi@ec.europa.eu, Edoardo.Ramalli@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
from ipywidgets import widgets, HTML, Layout
import ipyvuetify as v
import json

# Vois imports
from vois import download
from vois.vuetify import settings, page, selectImage, Button, UploadImage, UploadJson, sliderFloat, ColorPicker, switch, selectSingle, tabs, sortableList, dialogGeneric, iconButton

# Local imports
import mainPage
import PageConfigurator


# Panels dimensioning
LEFT_WIDTH = '30vw'


#####################################################################################################################################################
# mainPage Configurator
#####################################################################################################################################################
class MainConfigurator(page.page):


    # Initialization
    def __init__(self, output, onclose=None, leftWidth=LEFT_WIDTH, **kwargs):
        super().__init__('mainPage', 'Visual configurator', output, onclose=onclose, copyrighttext='European Commission - Joint Research Centre',
                         left_back=True, **kwargs)

        # Initialize member variables
        self.leftWidth = 'calc(%s - 2px)'%leftWidth

        self.spacerX = v.Html(tag='div', style_='width: 10px; height:  0px;')
        self.spacerY = v.Html(tag='div', style_='width:  0px; height: 10px;')
        self.spacer  = v.Html(tag='div', style_='width: 10px; height: 10px;')
        
        self.debug = widgets.Output()
        
        self.updatePageEnabled = True
        
        
    #################################################################################################################################################
    # Create the page and returns the card widget where the content of the page must be displayed
    #################################################################################################################################################
    def create(self):
        super().create()
        
        # Cards for the panels
        st = 'border-radius: 0px; border-color: %s; border-width: 1px; overflow: auto;'%settings.color_first
        self.main_width  = 'calc(100vw - %s)'%self.leftWidth
        self.main_height = self.height
        self.cardLeft   = v.Card(flat=True, style_=st + 'border-right-width: 0px;', outlined=True, width=self.leftWidth, min_width=self.leftWidth, max_width=self.leftWidth, height=self.height)
        self.cardMain   = v.Card(flat=True, style_=st + 'display: flex; justify-content: center; align-items: center;', outlined=True, width=self.main_width, height=self.main_height, color='#eeeeee')
        
        # Creation of the contents for the panels
        self.createMain()
        self.createLeft()
        
        # Compose the panels
        self.card.children = [ widgets.HBox([self.cardLeft, self.cardMain]) ]
        
        # Set the logo
        self.logowidth  = 70
        self.logoappurl = mainPage.getLocalImageURL('./graphics/main.png')
        
        return self.card

    
    #################################################################################################################################################
    # Create the content of the left panel
    #################################################################################################################################################
    def createLeft(self):
        
        self.labelwidth  = 160
        self.sliderwidth = 140
        
        
        maxwidth = 'calc(%s - 20px)'%LEFT_WIDTH
        self.controls = v.Card(flat=True, width=maxwidth, min_width=maxwidth, max_width=maxwidth, class_='pa-3 pt-1 ma-0', style_='overflow: auto;')
        
        netwidth = 'calc(%s - 58px)'%LEFT_WIDTH
        titwidth = 'calc(%s - 210px)'%LEFT_WIDTH
        
        self.upload = UploadImage.UploadImage(self.output, width=620)
        
        self.tftitle = v.TextField(label='Application title:', autofocus=False, v_model=self.main.title, dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3 mr-6")
        self.text_color = ColorPicker(dark=False, color=self.main.text_color, width=50, height=30, rounded=False, on_change=self.text_colorChange,  offset_x=True, offset_y=False, color_theory_popup=True)
        ct  = v.Card(flat=True, children=[self.tftitle], width=titwidth, min_width=titwidth, max_width=titwidth)
        cc  = v.Card(flat=True, children=[widgets.HBox([PageConfigurator.label('Text color:', color='black', width=66), self.text_color, self.text_color.ctpopup.draw()])], class_='pa-0 ma-0 mt-4')
        cr1 = v.Card(flat=True, children=[widgets.HBox([ct, cc])], width=netwidth, max_width=netwidth)
        
        self.tfsubtitle = v.TextField(label='Application subtitle:', autofocus=False, v_model=self.main.subtitle, dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3 mr-3")
        
        
        self.tftitle.on_event(   'change',      self.titleChange)
        self.tftitle.on_event(   'click:clear', self.titleClear)
        self.tfsubtitle.on_event('change',      self.subtitleChange)
        self.tfsubtitle.on_event('click:clear', self.subtitleClear)
        
        self.titlesizepercent = sliderFloat.sliderFloat(self.main.titlesizepercent, text='Title size percent:', minvalue=0, maxvalue=200, maxint=200, showpercentage=True, decimals=0,
                                                        labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_titlesizepercent)
        self.subtitlesizepercent = sliderFloat.sliderFloat(self.main.subtitlesizepercent, text='Subtitle size percent:', minvalue=0, maxvalue=200, maxint=200, showpercentage=True, decimals=0,
                                                           labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_subtitlesizepercent)
        
        self.titleshadow = switch.switch(self.main.titleshadow, 'Add shadow around application title', inset=True, dense=True, onchange=self.titleshadowChange)
        self.titleshadow_color = ColorPicker(dark=False, color=self.main.titleshadow_color, width=50, height=30, rounded=False, on_change=self.titleshadow_colorChange,  offset_x=True, offset_y=False, color_theory_popup=True)
        
        self.applogo_image_url  = Button('Select image for the application logo', color_selected=settings.color_first, dark=settings.dark_mode, 
                                         text_weight=450, on_click=self.applogo_image_load, width=netwidth, height=40,
                                         tooltip='Click to select an image to use as application logo on the title bar', selected=True, rounded=settings.button_rounded)
        self.slogowidth = sliderFloat.sliderFloat(self.main.applogo_widthpercent, text='Application logo width %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                 labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.logowidthChange)
        
        
        self.title_width   = sliderFloat.sliderFloat(self.main.titlebox_widthpercent, text='Title box width %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                     labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_titlebox_widthpercent)
        self.title_height  = sliderFloat.sliderFloat(self.main.titlebox_heightpercent, text='Title box height %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                     labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_titlebox_heightpercent)
        self.title_top     = sliderFloat.sliderFloat(self.main.titlebox_toppercent, text='Title box top %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                     labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_titlebox_toppercent)
        self.title_opacity = sliderFloat.sliderFloat(self.main.titlebox_opacity*100.0, text='Title box opacity %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                     labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_titlebox_opacity)
        self.title_border  = sliderFloat.sliderFloat(self.main.titlebox_border, text='Title box border (px):', minvalue=0.0, maxvalue=10.0, maxint=10, showpercentage=False, decimals=0,
                                                     labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_titlebox_border)
        
        self.background_image_url = Button('Select image for the application background', color_selected=settings.color_first, dark=settings.dark_mode, 
                                           text_weight=450, on_click=self.background_image_load, width=netwidth, height=40,
                                           tooltip='Click to select an image to use as application background', selected=True, rounded=settings.button_rounded)
        
        images = [{ "name": '%d'%x, "image": 'https://jeodpp.jrc.ec.europa.eu/services/shared/wallpapers/%d.jpg'%x, "max_width": 150, "max_height": 100 } for x in range(60)]
        self.background_image = selectImage.selectImage(images=images, selection=55, onchange=self.background_image_selected, width='380px',
                                                        label='Please select a stock image from the list', max_height=100, dense=True, outlined=True, clearable=True, margins="ma-0 mt-2 mr-1")

        self.filter_enabled = True
        self.filter_reset = Button('Reset all filters', color_selected=settings.color_first, dark=settings.dark_mode, 
                                    text_weight=450, on_click=self.onFilterReset, width=190, height=40,
                                    tooltip='Click to reset all filters for the background image', selected=True, rounded=settings.button_rounded)
        
        
        self.filter_blur = selectSingle.selectSingle('Blur:', ['0', '1', '2', '3', '4', '5', '6'], selection='0', width=80, clearable=False, onchange=self.onFilterChange)
        
        self.filter_grayscale = sliderFloat.sliderFloat(0, text='Grayscale:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                        labelwidth=self.labelwidth-70, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.onFilterChange)
        
        self.filter_brightness = sliderFloat.sliderFloat(100, text='Brightness:', minvalue=0.0, maxvalue=200.0, maxint=200, showpercentage=True, decimals=0,
                                                         labelwidth=self.labelwidth-70, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.onFilterChange)

        self.filter_contrast = sliderFloat.sliderFloat(100, text='Contrast:', minvalue=0.0, maxvalue=200.0, maxint=200, showpercentage=True, decimals=0,
                                                       labelwidth=self.labelwidth-70, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.onFilterChange)

        self.filter_saturate = sliderFloat.sliderFloat(100, text='Saturate:', minvalue=0.0, maxvalue=200.0, maxint=200, showpercentage=True, decimals=0,
                                                      labelwidth=self.labelwidth-70, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.onFilterChange)
        
        self.filter_hue = sliderFloat.sliderFloat(0, text='Hue:', minvalue=0.0, maxvalue=360.0, maxint=360, showpercentage=True, decimals=0,
                                                  labelwidth=self.labelwidth-70, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.onFilterChange)
        
        self.filter_opacity = sliderFloat.sliderFloat(100, text='Opacity:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                      labelwidth=self.labelwidth-70, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.onFilterChange)
        
        self.filter_sepia = sliderFloat.sliderFloat(0, text='Sepia:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                      labelwidth=self.labelwidth-70, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.onFilterChange)
        
        
        self.buttonbox_width  = sliderFloat.sliderFloat(self.main.buttonbox_widthpercent, text='Button box width %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                        labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_buttonbox_widthpercent)
        self.buttonbox_height = sliderFloat.sliderFloat(self.main.buttonbox_heightpercent, text='Button box height %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                        labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_buttonbox_heightpercent)
        self.buttonbox_top    = sliderFloat.sliderFloat(self.main.buttonbox_toppercent, text='Button box top %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                        labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_buttonbox_toppercent)

        
        self.button_width  = sliderFloat.sliderFloat(self.main.button_widthpercent, text='Buttons width %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                        labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_button_width)
        self.button_height = sliderFloat.sliderFloat(self.main.button_heightpercent, text='Buttons height %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                        labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_button_height)
        self.button_elevation = sliderFloat.sliderFloat(self.main.button_elevation, text='Buttons elevation:', minvalue=0.0, maxvalue=10.0, maxint=10, showpercentage=False, decimals=0,
                                                        labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_button_elevation)
        self.button_opacity = sliderFloat.sliderFloat(self.main.titlebox_opacity*100.0, text='Buttons opacity %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                      labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_button_opacity)
        self.button_titlesize = sliderFloat.sliderFloat(float(self.main.button_titlesize.replace('vh','')), text='Buttons title size:', minvalue=0.1, maxvalue=4.0, maxint=39, showpercentage=False, decimals=1,
                                                        labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_button_titlesize)
        self.button_subtitlesize = sliderFloat.sliderFloat(float(self.main.button_subtitlesize.replace('vh','')), text='Buttons subtitle size:', minvalue=0.1, maxvalue=4.0, maxint=39, showpercentage=False, decimals=1,
                                                           labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_button_subtitlesize)
        self.button_radius = sliderFloat.sliderFloat(float(self.main.button_radius.replace('px','')), text='Button radius (px):', minvalue=0.0, maxvalue=20.0, maxint=20, showpercentage=False, decimals=0,
                                                     labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_button_radius)
        
        self.blist = sortableList.sortableList(items=self.main.buttons, dark=False, allowNew=True, itemNew=self.buttonNew, itemContent=self.buttonDisplay, onchange=self.AssignOrderToButtons,
                                               onadded=self.buttonsUpdate, onremoved=self.buttonsUpdate, onmovedown=self.buttonsUpdate, onmoveup=self.buttonsUpdate)
        
        self.vois_show = switch.switch(self.main.vois_show, 'Show vois credits', inset=True, dense=True, onchange=self.vois_showChange)
        self.vois_opacity = sliderFloat.sliderFloat(self.main.vois_opacity*100.0, text='vois opacity:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                    labelwidth=90, sliderwidth=100, resetbutton=True, showtooltip=True, onchange=self.vois_opacityChange)
        if not self.main.vois_show: self.vois_opacity.slider.disabled = True

        self.tfcredits = v.TextField(label='Credits text:', autofocus=False, v_model=self.main.credits, dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3 mr-3")
        self.tfcredits.on_event('change',      self.creditsChange)
        self.tfcredits.on_event('click:clear', self.creditsClear)
        
        self.creditslogo_url  = Button('Select image for credits box', color_selected=settings.color_first, dark=settings.dark_mode, 
                                       text_weight=450, on_click=self.creditslogo_image_load, width=netwidth, height=40,
                                       tooltip='Click to select an image to use as credits logo on the bottom bar', selected=True, rounded=settings.button_rounded)
        
        self.credits_width   = sliderFloat.sliderFloat(self.main.creditbox_widthpercent, text='Credits box width %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                       labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_creditbox_widthpercent)
        self.credits_height  = sliderFloat.sliderFloat(self.main.creditbox_heightpercent, text='Credits box height %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                       labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_creditbox_heightpercent)
        self.credits_top     = sliderFloat.sliderFloat(self.main.creditbox_toppercent, text='Credits box top %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                       labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_creditbox_toppercent)
        self.credits_opacity = sliderFloat.sliderFloat(self.main.creditbox_opacity*100.0, text='Credits box opacity %:', minvalue=0.0, maxvalue=100.0, maxint=100, showpercentage=True, decimals=0,
                                                       labelwidth=self.labelwidth, sliderwidth=self.sliderwidth, resetbutton=True, showtooltip=True, onchange=self.on_creditbox_opacity)
        
        self.tfdisclaimer = v.TextField(label='Disclaimer:', autofocus=False, v_model=self.main.disclaimer, dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3 mr-3")
        self.tfdisclaimer.on_event(   'change',      self.disclaimerChange)
        self.tfdisclaimer.on_event(   'click:clear', self.disclaimerClear)
        
        
        # Tabs to group controls
        self.card_title      = v.Card(flat=True)
        self.card_background = v.Card(flat=True)
        self.card_buttons    = v.Card(flat=True)
        self.card_credits    = v.Card(flat=True)
        self.card_vois       = v.Card(flat=True)
        self.tabsView = tabs.tabs(0, ['Title', 'Background', 'Buttons', 'Credits', 'Vois'],
                                  contents=[self.card_title, self.card_background, self.card_buttons, self.card_credits, self.card_vois],
                                  dark=False, onchange=None, row=True)

        
        self.buttOpen  = iconButton.iconButton(icon='mdi-folder-open',  onclick=self.onOpen,  tooltip='Load state from file',                 margins='pa-0 ma-0 mt-3 mr-2', color=settings.color_first)
        self.buttSave  = iconButton.iconButton(icon='mdi-content-save', onclick=self.onSave,  tooltip='Save current state to file',           margins='pa-0 ma-0 mt-3 mr-2', color=settings.color_first)
        self.buttCode  = iconButton.iconButton(icon='mdi-file-code',    onclick=self.onCode,  tooltip='Generate and download code in Python', margins='pa-0 ma-0 mt-3 mr-2', color=settings.color_first)
        self.buttReset = iconButton.iconButton(icon='mdi-backspace',    onclick=self.onReset, tooltip='Reset main page to default state',     margins='pa-0 ma-0 mt-3 mr-2', color=settings.color_first)
        
        self.card_title.children = [widgets.VBox([
            self.spacerY,
            self.spacerY,
            cr1,
            self.tfsubtitle,
            self.titlesizepercent.draw(),
            self.subtitlesizepercent.draw(),
            self.applogo_image_url,
            self.slogowidth.draw(),
            self.title_width.draw(),
            self.title_height.draw(),
            self.title_top.draw(),
            self.title_opacity.draw(),
            self.title_border.draw(),
            self.titleshadow.draw(),
            widgets.HBox([PageConfigurator.label('Shadow color:', color='black'), self.titleshadow_color, self.titleshadow_color.ctpopup.draw()])
        ])]
        
        self.card_background.children = [widgets.VBox([
            self.spacerY,
            self.spacerY,
            self.background_image_url,
            self.spacerY,
            self.spacerY,
            self.background_image,
            self.spacerY,
            self.spacerY,
            widgets.HBox([self.filter_reset, self.spacerX, self.spacerX, self.spacerX, self.filter_blur.draw()]), 
            self.filter_grayscale.draw(),
            self.filter_brightness.draw(),
            self.filter_contrast.draw(),
            self.filter_saturate.draw(),
            self.filter_hue.draw(),
            self.filter_sepia.draw(),
            self.filter_opacity.draw()
        ])]
        
        
        self.card_buttons.children = [widgets.VBox([
            self.spacerY,
            self.spacerY,
            self.buttonbox_width.draw(),
            self.buttonbox_height.draw(),
            self.buttonbox_top.draw(),
            self.button_width.draw(),
            self.button_height.draw(),
            self.button_elevation.draw(),
            self.button_opacity.draw(),
            self.button_titlesize.draw(),
            self.button_subtitlesize.draw(),
            self.button_radius.draw(),
            PageConfigurator.label('Buttons:'),
            self.blist.draw()
        ])]
        
        
        self.card_credits.children = [widgets.VBox([
            self.spacerY,
            self.spacerY,
            self.tfcredits,
            self.spacerY,
            self.creditslogo_url,
            self.spacerY,
            self.spacerY,
            self.tfdisclaimer,
            self.spacerY,
            self.spacerY,
            self.credits_width.draw(),
            self.credits_height.draw(),
            self.credits_top.draw(),
            self.credits_opacity.draw()
        ])]
        
        
        self.card_vois.children = [widgets.VBox([
            self.spacerY,
            self.spacerY,
            self.vois_show.draw(),
            self.spacerY,
            self.vois_opacity.draw()
        ])]
        
        self.controls.children = [ widgets.VBox([
            widgets.HBox([self.buttOpen.draw(), self.buttSave.draw(), self.buttCode.draw(), self.buttReset.draw()]),
            self.tabsView.draw()
        ])]
        
        self.cardLeft.children = [self.controls]

    
    
    #################################################################################################################################################
    # Create the content of the Main panel
    #################################################################################################################################################
    def createMain(self):

        self.main = mainPage.mainPage(background_image=55)
        self.updatePreview()
        
        
    #################################################################################################################################################
    # Open, Save Reset buttons
    #################################################################################################################################################
        
    # Load state from file
    def onOpen(self):
        uj = UploadJson.UploadJson(self.output, onOK=self.onSelectedState, required_attributes=['title', 'subtitle'], attributes_width=80)
        uj.show()
        
        
    # Called when the UploadJson dialog-box is closed with the OK button
    def onSelectedState(self, state):
        self.main.fromJson(state)
        self.updateWidgets()
        self.updatePreview()
            
            
        
    # Save current state to file
    def onSave(self):
        self.saveFilename = v.TextField(label='File name:', autofocus=True, v_model=self.main.title, dense=False, color=settings.color_first, clearable=False, class_="pa-0 ma-0 ml-3 mt-3 mr-3")
        dialogGeneric.dialogGeneric(title='Save and download current state...' , on_ok=self.onDoSaveState, 
                                    text='   ', color=settings.color_first, dark=settings.dark_mode,
                                    titleheight=40, width=600,
                                    show=True, addclosebuttons=True, addokcancelbuttons=True,
                                    fullscreen=False, content=[self.saveFilename], output=self.output)
        
    
    # Effective save and download of the current state
    def onDoSaveState(self):
        
        filename = self.saveFilename.v_model
        if filename[-5:] != '.json':
            filename += '.json'
        
        state = self.main.toJson()
    
        # Convert to string and download
        txt = json.dumps(state, indent=4)
        download.downloadText(txt, fileName=filename)
        
                
    # Save Python code
    def onCode(self):
        self.saveFilename = v.TextField(label='File name:', autofocus=True, v_model=self.main.title, dense=False, color=settings.color_first, clearable=False, class_="pa-0 ma-0 ml-3 mt-3 mr-3")
        dialogGeneric.dialogGeneric(title='Save and download Python code...' , on_ok=self.onDoSaveCode, 
                                    text='   ', color=settings.color_first, dark=settings.dark_mode,
                                    titleheight=40, width=600,
                                    show=True, addclosebuttons=True, addokcancelbuttons=True,
                                    fullscreen=False, content=[self.saveFilename], output=self.output)

    
    # Effective save and download of the Python code
    def onDoSaveCode(self):
        
        self.onDoSaveState()   # Saves <filename>.json
        
        with self.debug:
            jsonfile = self.saveFilename.v_model
            if jsonfile[-5:] != '.json':
                jsonfile += '.json'

            notebookfile = self.saveFilename.v_model
            notebookfile += '.ipynb'

            scallbacks = ''
            for b in self.main.buttons:
                if 'onclick' in b and isinstance(b['onclick'], str) and len(b['onclick']) > 0 and not ' ' in b['onclick']:
                    scallbacks += '''
def %s(*args):
    dialogMessage.dialogMessage(title='Clicked on button %s', titleheight=36,
                                text='Change this lines with your code',
                                addclosebuttons=True, show=True, width=400, output=output)
    
'''%(b['onclick'],b['title'])

            # Code generation
            txt = '''from vois.vuetify import settings
settings.dark_mode      = True
settings.color_first    = '#0d856d'
settings.color_second   = '#a0dcd0'
settings.button_rounded = False

from ipywidgets import widgets, HTML, Layout
from IPython.display import display
import json

from vois import cssUtils
from vois.vuetify import mainPage, dialogMessage

output = widgets.Output(layout=Layout(width='0px', height='0px'))
display(output)

cssUtils.allSettings(output)
cssUtils.switchFontSize(output,14)

m = mainPage.mainPage()

%s

with open('%s') as f:
    j = json.load(f)
    m.fromJson(j)
    
for b in m.buttons:
    if 'onclick' in b and isinstance(b['onclick'], str) and len(b['onclick']) > 0 and not ' ' in b['onclick']:
        b['onclick'] = globals()[b['onclick']]
    
m.open()'''%(scallbacks, jsonfile)

            lines = txt.split('\n')
            lines = ",\n".join(["\"%s\\n\""%x for x in lines])

            txt = '''{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "65d32196-829e-41b9-a4e4-bdb6381998e9",
   "metadata": {},
   "outputs": [],
   "source": [
    %s
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python [interapro_env]",
   "language": "python",
   "name": "conda-env-interapro_env-py"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.6"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
'''%lines
        
            # Download .ipynb file
            download.downloadText(txt, fileName=notebookfile)
            
            
            
    # Reset page to initial state
    def onReset(self):
        self.main = mainPage.mainPage(background_image=55)
        self.updateWidgets()
        self.updatePreview()
        
        
        
    #################################################################################################################################################
    # Update widgets reading members of self.main
    #################################################################################################################################################
    def updateWidgets(self):
        
        self.updatePageEnabled = False
        
        self.tftitle.v_model    = self.main.title
        self.text_color.color   = self.main.text_color
        self.tfsubtitle.v_model = self.main.subtitle
        
        self.titlesizepercent.value    = self.main.titlesizepercent
        self.subtitlesizepercent.value = self.main.subtitlesizepercent

        self.titleshadow.value       = self.main.titleshadow
        self.titleshadow_color.color = self.main.titleshadow_color
        
        self.slogowidth.value = self.main.applogo_widthpercent
        
        self.title_width.value    = self.main.titlebox_widthpercent
        self.title_height.value   = self.main.titlebox_heightpercent
        self.title_top.value      = self.main.titlebox_toppercent
        self.title_opacity.value  = self.main.titlebox_opacity*100.0
        self.title_border.value   = self.main.titlebox_border
        
        if isinstance(self.main.background_image, int):
            self.background_image.value = self.main.background_image
        else:
            self.background_image.value = -1

        self.filter_blur.value       = '0'
        self.filter_grayscale.value  = 0
        self.filter_brightness.value = 100
        self.filter_contrast.value   = 100
        self.filter_saturate.value   = 100
        self.filter_hue.value        = 0
        self.filter_opacity.value    = 100
        self.filter_sepia.value      = 0
                                                    
        filters = self.main.background_filter.strip().split(' ')
        for f in filters:
            if '(' in f and ')' in f:
                if 'blur' in f:
                    s = f.replace('blur','').replace('(','').replace(')','').strip().replace('px','')
                    if len(s) > 0:
                        self.filter_blur.value = str(int(s))

                if 'grayscale' in f:
                    s = f.replace('grayscale','').replace('(','').replace(')','').strip().replace('%','')
                    if len(s) > 0:
                        self.filter_grayscale.value = int(s)

                if 'brightness' in f:
                    s = f.replace('brightness','').replace('(','').replace(')','').strip().replace('%','')
                    if len(s) > 0:
                        self.filter_brightness.value = int(s)

                if 'contrast' in f:
                    s = f.replace('contrast','').replace('(','').replace(')','').strip().replace('%','')
                    if len(s) > 0:
                        self.filter_contrast.value = int(s)

                if 'saturate' in f:
                    s = f.replace('saturate','').replace('(','').replace(')','').strip().replace('%','')
                    if len(s) > 0:
                        self.filter_saturate.value = int(s)

                if 'hue-rotate' in f:
                    s = f.replace('hue-rotate','').replace('(','').replace(')','').strip().replace('deg','')
                    if len(s) > 0:
                        self.filter_hue.value = int(s)

                if 'sepia' in f:
                    s = f.replace('sepia','').replace('(','').replace(')','').strip().replace('%','')
                    if len(s) > 0:
                        self.filter_sepia.value = int(s)

                if 'opacity' in f:
                    s = f.replace('opacity','').replace('(','').replace(')','').strip().replace('%','')
                    if len(s) > 0:
                        self.filter_opacity.value = int(s)
            
        
        self.buttonbox_width.value     = self.main.buttonbox_widthpercent
        self.buttonbox_height.value    = self.main.buttonbox_heightpercent
        self.buttonbox_top.value       = self.main.buttonbox_toppercent
        
        self.button_width.value        = self.main.button_widthpercent
        self.button_height.value       = self.main.button_heightpercent
        self.button_elevation.value    = self.main.button_elevation
        self.button_opacity.value      = self.main.titlebox_opacity*100.0
        self.button_titlesize.value    = float(self.main.button_titlesize.replace('vh',''))
        self.button_subtitlesize.value = float(self.main.button_subtitlesize.replace('vh',''))
        self.button_radius.value       = float(self.main.button_radius.replace('px',''))
        
        self.blist.items = self.main.buttons
        
        self.vois_show.value = self.main.vois_show
        self.vois_opacity.value = self.main.vois_opacity*100.0
        if not self.main.vois_show: self.vois_opacity.slider.disabled = True

        self.tfcredits.value = self.main.credits
        
        self.credits_width.value   = self.main.creditbox_widthpercent
        self.credits_height.value  = self.main.creditbox_heightpercent
        self.credits_top.value     = self.main.creditbox_toppercent
        self.credits_opacity.value = self.main.creditbox_opacity*100.0
        
        self.tfdisclaimer.value = self.main.disclaimer
        
        self.updatePageEnabled = True
        
        
    #################################################################################################################################################
    # Management of events in widgets
    #################################################################################################################################################
        
    # Update the preview
    def updatePreview(self):
        self.cardMain.children = [self.main.preview()]
        
        
    
    # Change application title
    def titleChange(self, *args):
        if self.updatePageEnabled:
            if self.tftitle.v_model is None: name = ''
            else:                            name = self.tftitle.v_model
            self.main.title = name
            self.updatePreview()

    def titleClear(self, *args):
        if self.updatePageEnabled:
            self.main.title = ''
            self.updatePreview()

        
    # Change application subtitle
    def subtitleChange(self, *args):
        if self.updatePageEnabled:
            if self.tfsubtitle.v_model is None: name = ''
            else:                               name = self.tfsubtitle.v_model
            self.main.subtitle = name
            self.updatePreview()

    def subtitleClear(self, *args):
        if self.updatePageEnabled:
            self.main.subtitlesizepercent = ''
            self.updatePreview()

        
    def on_titlesizepercent(self, value):
        if self.updatePageEnabled:
            self.main.titlesizepercent = value
            self.updatePreview()
    
    def on_subtitlesizepercent(self, value):
        if self.updatePageEnabled:
            self.main.subtitlesizepercent = value
            self.updatePreview()

    # Change of the text_color
    def text_colorChange(self):
        if self.updatePageEnabled:
            self.main.text_color = self.text_color.color
            self.updatePreview()

        
    # titleshadow flag
    def titleshadowChange(self, flag):
        if self.updatePageEnabled:
            self.main.titleshadow = flag
            self.updatePreview()
        
    # titleshadow_color
    def titleshadow_colorChange(self,):
        if self.updatePageEnabled:
            self.main.titleshadow_color = self.titleshadow_color.color
            self.updatePreview()
    
    
    # Selection of the application image to load from the local machine
    def applogo_image_loaded(self, imageurl):
        if self.updatePageEnabled:
            self.main.applogo_url = imageurl
            self.updatePreview()

    # Click on the button to open the dialog to select the the application image
    def applogo_image_load(self, *args):
        if self.updatePageEnabled:
            self.upload.title = 'Select image for the application logo'
            self.upload.onOK  = self.applogo_image_loaded
            self.upload.show()

        
    # Change logo width
    def logowidthChange(self, width):
        if self.updatePageEnabled:
            self.main.applogo_widthpercent = int(width)
            self.updatePreview()
        
    # titlebox_widthpercent   
    def on_titlebox_widthpercent(self, value):
        if self.updatePageEnabled:
            self.main.titlebox_widthpercent = int(value)
            self.updatePreview()

    # titlebox_heightpercent
    def on_titlebox_heightpercent(self, value):
        if self.updatePageEnabled:
            self.main.titlebox_heightpercent = int(value)
            self.updatePreview()
        
    # titlebox_toppercent
    def on_titlebox_toppercent(self, value):
        if self.updatePageEnabled:
            self.main.titlebox_toppercent = int(value)
            self.updatePreview()
        
    # titlebox_opacity
    def on_titlebox_opacity(self, value):
        if self.updatePageEnabled:
            self.main.titlebox_opacity = value/100.0
            self.updatePreview()

    # titlebox_widthpercent   
    def on_titlebox_border(self, value):
        if self.updatePageEnabled:
            self.main.titlebox_border = int(value)
            self.updatePreview()
        
        
    # Selection of the backgound image to load from the local machine
    def background_image_loaded(self, imageurl):
        if self.updatePageEnabled:
            self.main.background_image = imageurl
            self.background_image.value = -1
            self.updatePreview()

    # Click on the button to open the dialog to select the the background image
    def background_image_load(self, *args):
        if self.updatePageEnabled:
            self.upload.title = 'Select image for the application background'
            self.upload.onOK  = self.background_image_loaded
            self.upload.show()
        
        
    # Selection of a stock background image
    def background_image_selected(self):
        if self.updatePageEnabled:
            if self.background_image.value >= 0:
                self.main.background_image = self.background_image.value
                self.updatePreview()
        
        
    # Filter reset
    def onFilterReset(self):
        if self.updatePageEnabled:
            self.filter_enabled = False
            self.filter_blur.value       = '0'
            self.filter_grayscale.value  = 0
            self.filter_brightness.value = 100
            self.filter_contrast.value   = 100
            self.filter_saturate.value   = 100
            self.filter_hue.value        = 0
            self.filter_opacity.value    = 100
            self.filter_sepia.value      = 0
            self.filter_enabled = True
            self.onFilterChange()
        
        
    # Filter change
    def onFilterChange(self, *args):
        if self.updatePageEnabled:
            if self.filter_enabled:
                filter = ''

                n = int(self.filter_blur.value)
                if n > 0: filter += 'blur(%dpx) '%n

                n = int(self.filter_grayscale.value)
                if n > 0: filter += 'grayscale(%d%%) '%n

                n = int(self.filter_brightness.value)
                if n != 100: filter += 'brightness(%d%%) '%n

                n = int(self.filter_contrast.value)
                if n != 100: filter += 'contrast(%d%%) '%n

                n = int(self.filter_saturate.value)
                if n != 100: filter += 'saturate(%d%%) '%n

                n = int(self.filter_hue.value)
                if n != 0: filter += 'hue-rotate(%ddeg) '%n

                n = int(self.filter_sepia.value)
                if n != 0: filter += 'sepia(%d%%) '%n

                n = int(self.filter_opacity.value)
                if n != 100: filter += 'opacity(%d%%) '%n

                self.main.background_filter = filter
                self.updatePreview()
        
        
    # buttonbox_widthpercent   
    def on_buttonbox_widthpercent(self, value):
        if self.updatePageEnabled:
            self.main.buttonbox_widthpercent = int(value)
            self.updatePreview()

    # buttonbox_heightpercent
    def on_buttonbox_heightpercent(self, value):
        if self.updatePageEnabled:
            self.main.buttonbox_heightpercent = int(value)
            self.updatePreview()
        
    # buttonbox_toppercent
    def on_buttonbox_toppercent(self, value):
        if self.updatePageEnabled:
            self.main.buttonbox_toppercent = int(value)
            self.updatePreview()

        
    # button_widthpercent
    def on_button_width(self, value):
        if self.updatePageEnabled:
            self.main.button_widthpercent = int(value)
            self.updatePreview()

    # button_heightpercent
    def on_button_height(self, value):
        if self.updatePageEnabled:
            self.main.button_heightpercent = int(value)
            self.updatePreview()

    # button_elevation
    def on_button_elevation(self, value):
        if self.updatePageEnabled:
            self.main.button_elevation = int(value)
            self.updatePreview()

    # button_opacity
    def on_button_opacity(self, value):
        if self.updatePageEnabled:
            self.main.button_opacity = value/100.0
            self.updatePreview()

    # button_titlesize
    def on_button_titlesize(self, value):
        if self.updatePageEnabled:
            self.main.button_titlesize = '%fvh'%value
            self.updatePreview()

    # button_subtitlesize
    def on_button_subtitlesize(self, value):
        if self.updatePageEnabled:
            self.main.button_subtitlesize = '%fvh'%value
            self.updatePreview()
            self.button_subtitlesize.draw(),
            
    # button_radius
    def on_button_radius(self, value):
        if self.updatePageEnabled:
            self.main.button_radius = '%dpx'%int(value)
            self.updatePreview()

        
    # Change of vois_show flag
    def vois_showChange(self, flag):
        if self.updatePageEnabled:
            self.main.vois_show = flag
            self.vois_opacity.slider.disabled = not self.main.vois_show
            self.updatePreview()
        
        
    # Change vois opacity
    def vois_opacityChange(self, opacity):
        if self.updatePageEnabled:
            self.main.vois_opacity = opacity / 100.0
            self.updatePreview()

        
    # Change credits text
    def creditsChange(self, *args):
        if self.updatePageEnabled:
            if self.tfcredits.v_model is None: name = ''
            else:                              name = self.tfcredits.v_model
            self.main.credits = name
            self.updatePreview()

    def creditsClear(self, *args):
        if self.updatePageEnabled:
            self.main.credits = ''
            self.updatePreview()
        
        
    # Selection of the credits image to load from the local machine
    def creditslogo_image_loaded(self, imageurl):
        if self.updatePageEnabled:
            self.main.creditslogo_url = imageurl
            self.updatePreview()

    # Click on the button to open the dialog to select the the credits image
    def creditslogo_image_load(self, *args):
        self.upload.title = 'Select image for the credits image'
        self.upload.onOK  = self.creditslogo_image_loaded
        self.upload.show()

        
    # Change disclaimer text
    def disclaimerChange(self, *args):
        if self.updatePageEnabled:
            if self.tfdisclaimer.v_model is None: name = ''
            else:                                 name = self.tfdisclaimer.v_model
            self.main.disclaimer = name
            self.updatePreview()

    def disclaimerClear(self, *args):
        if self.updatePageEnabled:
            self.main.disclaimer = ''
            self.updatePreview()


    # creditbox_widthpercent   
    def on_creditbox_widthpercent(self, value):
        if self.updatePageEnabled:
            self.main.creditbox_widthpercent = int(value)
            self.updatePreview()

    # creditbox_heightpercent
    def on_creditbox_heightpercent(self, value):
        if self.updatePageEnabled:
            self.main.creditbox_heightpercent = int(value)
            self.updatePreview()
        
        
    # creditbox_toppercent
    def on_creditbox_toppercent(self, value):
        if self.updatePageEnabled:
            self.main.creditbox_toppercent = int(value)
            self.updatePreview()
        
        
    # creditbox_opacity
    def on_creditbox_opacity(self, value):
        if self.updatePageEnabled:
            self.main.creditbox_opacity = value/100.0
            self.updatePreview()
        

    # Add a new button
    def buttonNew(self):
        eb = EditButton(self.output, index=len(self.main.buttons), newButton=True, onOK=self.AddNewButton)
        return None
    
    def AddNewButton(self, b):
        self.blist.doAddItem(b)
        self.updatePreview()
        
        
    # Display a button
    def buttonDisplay(self, item, index):
        
        def UpdateButton(b):
            self.main.buttons[b['index']] = b
            self.blist.items = self.main.buttons
            self.updatePreview()
            
        def onclick(*args):
            eb = EditButton(self.output, index=item['index'], title=item['title'], subtitle=item['subtitle'], tooltip=item['tooltip'], image=item['image'], onclick=item['onclick'], argument=item['argument'], newButton=False, onOK=UpdateButton)
        
        self.AssignOrderToButtons()
        lab = PageConfigurator.label(item['title'], class_='pa-0 ma-0 ml-2 mb-2', color='black')
        lab.on_event('click', onclick)
        return [lab]
        
        
    # Added or removed a button
    def buttonsUpdate(self, *args):
        if self.updatePageEnabled:
            self.updatePreview()
        
        
    # Called at each order change oand each remove of a button
    def AssignOrderToButtons(self):
        i = 0
        for b in self.main.buttons:
            b['index'] = i
            i += 1
        
        
#####################################################################################################################################################
# Create a New button or Edit info on a button
#####################################################################################################################################################
class EditButton():
    
    def __init__(self, output, index, title='', subtitle='', tooltip='', image='', onclick=None, argument=None, onOK=None, newButton=True):
        
        self.onOK = onOK
        self.upload = UploadImage.UploadImage(output, width=620)

        self.index = index
        
        self.image = image

        self.tit    = v.TextField(label='Title:',          autofocus=True,  v_model=title,    dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3")
        self.subtit = v.TextField(label='Subtitle:',       autofocus=False, v_model=subtitle, dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3")
        self.ttip   = v.TextField(label='Tooltip:',        autofocus=False, v_model=tooltip,  dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3")
        self.oncl   = v.TextField(label='On click call:',  autofocus=False, v_model=onclick,  dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3")
        self.arg    = v.TextField(label='Argument:',       autofocus=False, v_model=argument, dense=True, color=settings.color_first, clearable=True, class_="pa-0 ma-0 mt-3")

        b = Button('Select image for the button', color_selected=settings.color_first, dark=settings.dark_mode, 
                   text_weight=450, on_click=self.onSelectImage, width=570, height=40,
                   tooltip='Click to select an image to display on the button', selected=True, rounded=settings.button_rounded)

        c = v.Card(flat=True, children=[widgets.VBox([self.tit, self.subtit, self.ttip, self.oncl, self.arg, b])], class_='pa-4 ma-0')

        if newButton: dialogtitle = 'New button'
        else:         dialogtitle = 'Edit button'
        dlg = dialogGeneric.dialogGeneric(title=dialogtitle,
                                          text='',
                                          show=True, addclosebuttons=True, width=600,
                                          addokcancelbuttons=True, on_ok=self.AddedButton,
                                          fullscreen=False, content=[c], output=output)
        
        return None
        
        
    def onImageLoaded(self, imageurl):
        self.image = imageurl


    def onSelectImage(self):
        self.upload.title = 'Select image for the button'
        self.upload.onOK  = self.onImageLoaded
        self.upload.show()

    def AddedButton(self):
        self.button = {'title':    self.tit.v_model,
                       'subtitle': self.subtit.v_model,
                       'tooltip':  self.ttip.v_model,
                       'image':    self.image,
                       'onclick':  self.oncl.v_model,
                       'argument': self.arg.v_model,
                       'index':    self.index}
        
        if self.onOK is not None:
            self.onOK(self.button)
        


"""IconButton to copy text to the clipboard."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
from IPython.display import HTML as ipyHTML
import ipyvuetify as v

from vois.vuetify import settings, iconButton
            
###########################################################################################################################################################################
# Icon button to copy a text to the clipboard
###########################################################################################################################################################################
class IconClipboard(v.Html):
    """
    IconButton to copy text to the clipboard.
        
    Parameters
    ----------
    output : ipywidgets.Output
        Output widget to be used to execute javacript code to copy the text to the clipboard
    text : str, optional
        Text to be copied to the clipboard (default is '', it can be changed using the 'text' property)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    outlined : bool, optional
        If True applies a thin border to the button (default is False)
    rounded : bool, optional
        If True the shape of the button is rounded (default is True)
    width : str, optional
        Width of the widget (default is None)
    large : bool, optional
        If True makes the button large (default is False)
    small : bool, optional
        If True makes the button small (default is False)
    x_large : bool, optional
        If True makes the button extra-large (default is False)
    x_small : bool, optional
        If True makes the button extra-small (default is False)
    margins : str, optional
        Marging apply to the button (default is 'pa-0 ma-0')
    tooltip : str, optional
        Tooltip text for the button (default is 'Click to copy text to the clipboard')

    Example
    -------
    Creation and display of an icon button to copy some text content to the clipboard::
        
        from vois.vuetify import IconClipboard
        from IPython.display import display
        
        b = IconClipboard()
        display(b)
        
        b.text = 'Text to copy to the clipboard'
   """
    
    def __init__(self,
                 output,
                 text='',
                 color=None,
                 outlined=False,
                 rounded=True,
                 width=None,
                 large=False,
                 small=False,
                 x_large=False,
                 x_small=False,
                 margins='pa-0 ma-0',
                 tooltip='Click to copy text to the clipboard',
                 **kwargs
                ):
        
        super().__init__(**kwargs)
        
        self.output = output
        
        self._text = text

        self._color = color
        if self._color is None:
            self._color = settings.color_first
        
        self.b = iconButton.iconButton(icon='mdi-content-copy', color=self._color, onclick=self.onclick, tooltip=tooltip,
                                       outlined=outlined, rounded=rounded, width=width,
                                       large=large, small=small, x_large=x_large, x_small=x_small, margins=margins, disabled=len(self._text)==0)

        self.tag = 'div'
        self.children = [self.b.draw()]

        
    # "old" draw
    def draw(self):
        return self
        

    # Manage click event
    def onclick(self):
        with self.output:
            display(ipyHTML('''
<script>
navigator.clipboard.writeText("%s");
//navigator.clipboard.readText().then((clipText) => (console.log("clipboard: "+clipText)))
</script>'''%self._text))
    
    
    @property
    def text(self):
        return self._text
        
    @text.setter
    def text(self, t):
        self._text = t.replace("\n","\\"+"n")   # V.I. Need to have two separate character "\" + "n" if a text with newline is copied to the clipboard!!!
        
        self.b.disabled = len(self._text)==0



"""Example of resizable SVG drawing to use inside a Content"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
from ipywidgets import widgets, HTML, Layout
import ipyvuetify as v
import pandas as pd
import random

# Vois imports
from vois import svgMap
from vois.vuetify import settings, palettePickerEx
from vois.templates import PageConfigurator


#####################################################################################################################################################
# Example of resizable SVG drawing to use inside a Content
#####################################################################################################################################################
class SVGdrawing():
    
    def __init__(self,
                 width='30vw',        # Dimensions of the overall card displaying the SVG
                 height='40vh',
                 color_first=None,    # Main color
                 color_second=None,   # Secondary color
                 dark=None,           # Dark flag
                 **kwargs):
        
        self._width  = width
        self._height = height
        
        self._svg = ''
        self.customsvg = False   # True if user passed its own SVG
        
        # Colors of the configuration widgets
        self._color_first = color_first
        if self._color_first is None:
            self._color_first = settings.color_first
        
        self._color_second = color_second
        if self._color_second is None:
            self._color_second = settings.color_second
            
        self._dark = dark
        if self._dark is None:
            self._dark = settings.dark_mode

        # Create the card that will contain the SVG drawing
        self.card = v.Card(flat=True, tile=True, width=self._width, height=self._height,
                           style_='overflow: hidden;', class_='d-flex align-center justify-center')  # The content of the card is centered horizontally and vertically
        
        
        # Creation of the SVG displaying the map of Europe
        self.codes_selected = []
        self.df = pd.DataFrame(columns=['iso2code','value','label'])
        for code in svgMap.country_codes:
            value = random.randint(0,100)
            record = {'iso2code': code, 'value': value, 'label': '%d' % value }
            self.df.loc[len(self.df)] = record
            
            selected = random.randint(0,100) >= 90
            if selected:
                self.codes_selected.append(code)
        
        # Palette picker
        self.pp = None
        self.pp = palettePickerEx.palettePickerEx(family='sequential', value='Viridis', interpolate=True, show_interpolate_switch=False, clearable=False, width=200,
                                                  onchange=self.updateChart, show_opacity_slider=False, onchangeOpacity=None)
        
        self.updateChart()
        
        
    # Draw th widget
    def draw(self):
        return self.card
        
        
    # Updte the chart
    def updateChart(self, *args):
        if self.pp is not None and self.pp.colors is not None and len(self.pp.colors) > 1:
            
            if not self.customsvg:
                self._svg = svgMap.svgMapEurope(self.df,
                                                code_column='iso2code',
                                                width=self._width,
                                                height=self._height,
                                                stdevnumber=1.5, 
                                                colorlist=self.pp.colors,
                                                stroke_width=4.0,
                                                stroke_selected=self._color_first,
                                                onhoverfill='#f8bd1a',
                                                codes_selected=self.codes_selected,
                                                legendtitle='Legent title',
                                                legendunits='KTOE per 100K inhabit.')

            # Display the SVG inside the card
            self.card.children = [HTML(self._svg)]
        
        
    # Configure the SVG drawing
    def configure(self):
        return v.Card(flat=True, class_='pa-2 ma-0', children=[widgets.VBox([PageConfigurator.label('SVG drawing', color='black'), self.pp.draw()])])



    @property
    def content(self):
        return 'SVG Drawing'
        
    @content.setter
    def content(self, c):
        pass
    
    
    @property
    def width(self):
        return self._width
        
    @width.setter
    def width(self, w):
        self._width = w
        if self._svg is not None:
            self._svg = svgMap.svgChangeWidth(self._svg, self._width)
        self.card.width = w

        
    @property
    def height(self):
        return self._height
        
    @height.setter
    def height(self, h):
        self._height = h
        if self._svg is not None:
            self._svg = svgMap.svgChangeHeight(self._svg, self._height)
        self.card.height = h

        
    @property
    def color_first(self):
        return self._color_first
        
    @color_first.setter
    def color_first(self, color):
        self._color_first = color

        if self.pp is not None:
            self.pp.color = color
    
    
    @property
    def color_second(self):
        return self._color_second
        
    @color_second.setter
    def color_second(self, color):
        self._color_second = color
    
    
    @property
    def dark(self):
        return self._dark
        
    @dark.setter
    def dark(self, flag):
        self._dark = flag
    
    
    
    @property
    def familyname(self):
        return self.pp.familyname
    
    @familyname.setter
    def familyname(self, name):
        self.pp.familyname = name
    
    
    @property
    def palette(self):
        return self.pp.value
    
    @palette.setter
    def palette(self, name):
        self.pp.value = name
    
    
    @property
    def svg(self):
        return self._svg
    
    @svg.setter
    def svg(self, text):
        self._svg = text
        self.customsvg = True
        self.card.children = [HTML(self._svg)]
        
        
    @property
    def state(self):
        return {x: getattr(self, x) for x in ['content',
                                              #'width',      # Will inherit from content!!!
                                              #'height',
                                              'color_first',
                                              'color_second',
                                              'dark',
                                              'familyname',
                                              'palette'
                                             ]}
        
    @state.setter
    def state(self, statusdict):
        for key, value in statusdict.items():
            setattr(self, key, value)
    


"""CSS utility functions to ensure Voila' behaviour is coherent with JupyterLab."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from ipywidgets import HTML
from IPython.display import display
from IPython.display import HTML as ipyHTML
import ipyvuetify as v

from vois.vuetify import settings, iconButton


###########################################################################################################################################################################
# V.I. Avoid left margin in Output widgets inside dialog-boxes!!!
###########################################################################################################################################################################
def dialogBoxesLeftMargin(output=None):
    
    command = '''
<style>.jp-OutputPrompt {
    flex: 0 0 0 !important;
    display: inline !important;
    min-width: 0 !important;
}
</style>'''
    
    if output is None:
        display(HTML(command))
    else:
        with output:
            display(HTML(command))

            

###########################################################################################################################################################################
# Settings for popup when clicking on map: popup is correctly displayed also in Voila'
###########################################################################################################################################################################
def popupDisplay(output=None):

    command = '''
<style>
.leaflet-popup-tip {
    width: 17px !important;
    height: 17px !important;
    padding: 1px !important;
    margin: -10px auto 0 !important;
    -webkit-transform: rotate(45deg) !important;
    -moz-transform: rotate(45deg) !important;
    -ms-transform: rotate(45deg) !important;
    transform: rotate(45deg) !important;
}

.leaflet-popup-content-wrapper, .leaflet-popup-tip {
    background: #fff;
    color: #333;
    box-shadow: 0 3px 14px rgba(0,0,0,0.4);
}

.leaflet-popup-content-wrapper {
    padding: 1px !important;
    text-align: left !important;
    border-radius: 12px !important;
}
</style>'''

    if output is None:
        display(HTML(command))
    else:
        with output:
            display(HTML(command))

            
            
###########################################################################################################################################################################
# Change default font-size of the labels of the vuetify Switch widget
###########################################################################################################################################################################
def switchFontSize(output=None, size=14):
    
    command = '''
<style>.vuetify-styles .v-label {
    font-size: %dpx;
}
</style>'''%int(size)
    
    if output is None:
        display(HTML(command))
    else:
        with output:
            display(HTML(command))

    
###########################################################################################################################################################################
# Execute all settings by displaying all the <style> instructions
###########################################################################################################################################################################
def allSettings(output=None):
    dialogBoxesLeftMargin(output)
    popupDisplay(output)
    
    
    
###########################################################################################################################################################################
# Copy text to clipboard
###########################################################################################################################################################################
def copyToClipboard(txt, output=None):
    command = '''
<script>
    navigator.clipboard.writeText('%s').then(function() {
    console.log('Copying to clipboard was successful!');
  }, function(err) {
    console.error('Could not copy text: ', err);
  });    
</script>'''%(txt)
    
    if output is None:
        display(ipyHTML(command))
    else:
        with output:
            display(ipyHTML(command))



"""Simplified interface to the BDAP RASTERAPI."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Python imports
import json
import requests


#####################################################################################################################################################
# Python user-defined exceptions
#####################################################################################################################################################

# Bad answer from a BDAP HTTP(S) request
class InvalidBDAPAnswerException(Exception):
    "Raised when BDAP server fails to answer"

    def __init__(self, url, data=''):
        self.message = 'BDAP failed to correctly execute the command: ' + str(url)
        if len(data) > 0:
            self.message += '\nData: ' + str(data)
        super().__init__(self.message)    



#####################################################################################################################################################
# Query information on a raster file giving its server side full path
# Example: https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=INFO&filepath=/eos/jeodpp/data/base/Landcover/GLOBAL/UMD/GFC/VER1-7/Data/VRT/first/Hansen_GFC-2019-v1.7_first.vrt
#####################################################################################################################################################
def rasterInfo(filepath, request_stats=False, detailed_stats=False, NoneReplace=-1.0, nanReplace=-1.0):
    
    url = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=INFO&filepath=%s'%filepath
    
    # Calculate approximate stats without scanning all the pixels!
    if request_stats:
        url += '&STATS=1'
        
        if detailed_stats:
            url += '&DETAILED=1'
        
    req = requests.get(url)
    if req.status_code == 200:
        if len(req.text) > 0:
            #info = json.loads(req.text.replace('\'','"').replace('None','"None"'))
            info = json.loads(req.text.replace('\'','"').replace('None','%f'%NoneReplace).replace('nan','%f'%nanReplace).replace('-inf','%f'%nanReplace))
        else:
            info = {}    # In case the file is not existant!
    else:
        raise InvalidBDAPAnswerException(url=url)
        
    return info



#####################################################################################################################################################
# Save a layer definition on REDIS server and returns a procid string
#####################################################################################################################################################
def saveLayer(strjson):
    url = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=TOLAYER'

    req = requests.get(url, data=strjson)
    procid = None
    if req.status_code == 200:
        procid = str(req.text)
    else:
        raise InvalidBDAPAnswerException(url=url,data=strjson)
        
    return procid



#####################################################################################################################################################
# Calculate occurrencies of a raster band and returns a dictionary
# Example: https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=OCCURRENCIES&filepath=/eos/jeodpp/data/products/Landcover/EUROPE/EUCROPMAP/VER2022-1/Data/VRT/JRC_EUROCROPMAP2022_EU27_EPSG3035.vrt&band=1&epsg=3035&nodata=0.0&lon1=12.5&lon2=12.6&lat1=43.5&lat2=43.6&zoom=12
#####################################################################################################################################################
def rasterOccurrencies(filepath,
                       lonmin,
                       latmin,
                       lonmax,
                       latmax,
                       zoom,
                       band=1,
                       epsg=4326,
                       nodata=0.0):
    
    url = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=OCCURRENCIES&filepath=%s&band=%d&epsg=%d&nodata=%f&lon1=%f&lon2=%f&lat1=%f&lat2=%f&zoom=%d'%(filepath,band,epsg,nodata,lonmin,lonmax,latmin,latmax,zoom)
    
    req = requests.get(url)
    if req.status_code == 200:
        import ast
        return ast.literal_eval(req.text)
    else:
        raise InvalidBDAPAnswerException(url=url)

        
#####################################################################################################################################################
# Query raster. Read raster value at some points in geographic coordinates
#####################################################################################################################################################
def rasterQuery(filepath, band=1, lon=[], lat=[], NoneReplace=-1.0, nanReplace=-1.0, epsg=None):
    
    url = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=QUERY&filepath=%s&band=%d'%(filepath,band)
    
    if not epsg is None:
        url += '&epsg=%d'%int(epsg)
    
    j = { "lon": list(lon), "lat": list(lat) }
    strjson = json.dumps(j)
    
    req = requests.get(url, data=strjson)
    if req.status_code == 200:
        return json.loads(req.text.replace('\'','"').replace('None','%f'%NoneReplace).replace('nan','%f'%nanReplace).replace('-inf','%f'%nanReplace))
    else:
        raise InvalidBDAPAnswerException(url=url)



        